/*
 * map2.h - Easymesh-r2 CMDUs and TLVs definition in flat format.
 *
 * Copyright (C) 2021 IOPSYS Software Solutions AB. All rights reserved.
 *
 * See LICENSE file for license related information.
 *
 */
#ifndef MAP2_H
#define MAP2_H


#include <stdint.h>

#define CMDU_1905_ACK                                  0x8000
#define CMDU_AP_CAPABILITY_QUERY                       0x8001
#define CMDU_AP_CAPABILITY_REPORT                      0x8002
#define CMDU_POLICY_CONFIG_REQ                         0x8003
#define CMDU_CHANNEL_PREFERENCE_QUERY                  0x8004
#define CMDU_CHANNEL_PREFERENCE_REPORT                 0x8005
#define CMDU_CHANNEL_SELECTION_REQ                     0x8006
#define CMDU_CHANNEL_SELECTION_RESPONSE                0x8007
#define CMDU_OPERATING_CHANNEL_REPORT                  0x8008
#define CMDU_CLIENT_CAPABILITY_QUERY                   0x8009
#define CMDU_CLIENT_CAPABILITY_REPORT                  0x800a
#define CMDU_AP_METRICS_QUERY                          0x800b
#define CMDU_AP_METRICS_RESPONSE                       0x800c
#define CMDU_ASSOC_STA_LINK_METRICS_QUERY              0x800d
#define CMDU_ASSOC_STA_LINK_METRICS_RESPONSE           0x800e
#define CMDU_UNASSOC_STA_LINK_METRIC_QUERY             0x800f
#define CMDU_UNASSOC_STA_LINK_METRIC_RESPONSE          0x8010
#define CMDU_BEACON_METRICS_QUERY                      0x8011
#define CMDU_BEACON_METRICS_RESPONSE                   0x8012
#define CMDU_COMBINED_INFRA_METRICS                    0x8013
#define CMDU_CLIENT_STEERING_REQUEST                   0x8014
#define CMDU_CLIENT_STEERING_BTM_REPORT                0x8015
#define CMDU_CLIENT_ASSOC_CONTROL_REQUEST              0x8016
#define CMDU_STEERING_COMPLETED                        0x8017
#define CMDU_HIGHER_LAYER_DATA                         0x8018
#define CMDU_BACKHAUL_STEER_REQUEST                    0x8019
#define CMDU_BACKHAUL_STEER_RESPONSE                   0x801a

#define CMDU_CHANNEL_SCAN_REQUEST                      0x801b
#define CMDU_CHANNEL_SCAN_REPORT                       0x801c
#define CMDU_CAC_REQUEST                               0x8020
#define CMDU_CAC_TERMINATION                           0x8021
#define CMDU_CLIENT_DISASSOCIATION_STATS               0x8022
#define CMDU_ERROR_RESPONSE                            0x8024
#define CMDU_ASSOCIATION_STATUS_NOTIFICATION           0x8025
#define CMDU_TUNNELED                                  0x8026
#define CMDU_BACKHAUL_STA_CAPABILITY_QUERY             0x8027
#define CMDU_BACKHAUL_STA_CAPABILITY_REPORT            0x8028
#define CMDU_FAILED_CONNECTION                         0x8033

#define LAST_MAP_CMDU                                  CMDU_FAILED_CONNECTION
#define MAP_CMDU_TYPE_MAX                              LAST_MAP_CMDU


#define MAP_TLV_SUPPORTED_SERVICE                      0x80
#define MAP_TLV_SEARCHED_SERVICE                       0x81
#define MAP_TLV_AP_RADIO_IDENTIFIER                    0x82
#define MAP_TLV_AP_OPERATIONAL_BSS                     0x83
#define MAP_TLV_ASSOCIATED_CLIENTS                     0x84
#define MAP_TLV_AP_CAPABILITY                          0xA1
#define MAP_TLV_AP_RADIO_BASIC_CAPABILITIES            0x85
#define MAP_TLV_AP_HT_CAPABILITIES                     0x86
#define MAP_TLV_AP_VHT_CAPABILITIES                    0x87
#define MAP_TLV_AP_HE_CAPABILITIES                     0x88
#define MAP_TLV_STEERING_POLICY                        0x89
#define MAP_TLV_METRIC_REPORTING_POLICY                0x8A
#define MAP_TLV_CHANNEL_PREFERENCE                     0x8B
#define MAP_TLV_RADIO_OPERATION_RESTRICTION            0x8C
#define MAP_TLV_TRANSMIT_POWER_LIMIT                   0x8D
#define MAP_TLV_CHANNEL_SELECTION_RESPONSE             0x8E
#define MAP_TLV_OPERATING_CHANNEL_REPORT               0x8F
#define MAP_TLV_CLIENT_INFO                            0x90
#define MAP_TLV_CLIENT_CAPABILITY_REPORT               0x91
#define MAP_TLV_CLIENT_ASSOCIATION_EVENT               0x92
#define MAP_TLV_AP_METRIC_QUERY                        0x93
#define MAP_TLV_AP_METRICS                             0x94
#define MAP_TLV_STA_MAC_ADDRESS                        0x95
#define MAP_TLV_ASSOCIATED_STA_LINK_METRICS            0x96
#define MAP_TLV_UNASSOCIATED_STA_LINK_METRICS_QUERY    0x97
#define MAP_TLV_UNASSOCIATED_STA_LINK_METRICS_RESPONSE 0x98
#define MAP_TLV_BEACON_METRICS_QUERY                   0x99
#define MAP_TLV_BEACON_METRICS_RESPONSE                0x9A
#define MAP_TLV_STEERING_REQUEST                       0x9B
#define MAP_TLV_STEERING_BTM_REPORT                    0x9C
#define MAP_TLV_CLIENT_ASSOCIATION_CONTROL_REQUEST     0x9D
#define MAP_TLV_BACKHAUL_STEERING_REQUEST              0x9E
#define MAP_TLV_BACKHAUL_STEERING_RESPONSE             0x9F
#define MAP_TLV_HIGHER_LAYER_DATA                      0xA0
#define MAP_TLV_ASSOCIATED_STA_TRAFFIC_STATS           0xA2
#define MAP_TLV_ERROR_CODE                             0xA3

#define MAP_TLV_CHANNEL_SCAN_REPORTING_POLICY          0xA4
#define MAP_TLV_CHANNEL_SCAN_CAPABILITY                0xA5
#define MAP_TLV_CHANNEL_SCAN_REQ                       0xA6
#define MAP_TLV_CHANNEL_SCAN_RES                       0xA7
#define MAP_TLV_TIMESTAMP                              0xA8
#define MAP_TLV_CAC_REQ                                0xAD
#define MAP_TLV_CAC_TERMINATION                        0xAE
#define MAP_TLV_CAC_COMPLETION_REPORT                  0xAF
#define MAP_TLV_CAC_STATUS_REPORT                      0xB1
#define MAP_TLV_CAC_CAPABILITY                         0xB2
#define MAP_TLV_MULTIAP_PROFILE                        0xB3
#define MAP_TLV_PROFILE2_AP_CAPABILITY                 0xB4
#define MAP_TLV_DEFAULT_8021Q_SETTINGS                 0xB5
#define MAP_TLV_TRAFFIC_SEPARATION_POLICY              0xB6
#define MAP_TLV_PROFILE2_ERR_CODE                      0xBC
#define MAP_TLV_AP_RADIO_ADV_CAPABILITY                0xBE
#define MAP_TLV_ASSOCIATION_STATUS_NOTIF               0xBF
#define MAP_TLV_SOURCE_INFO                            0xC0
#define MAP_TLV_TUNNELED_MSG_TYPE                      0xC1
#define MAP_TLV_TUNNELED                               0xC2
#define MAP_TLV_PROFILE2_STEERING_REQ                  0xC3
#define MAP_TLV_UNSUCCESS_ASSOCIATION_POLICY           0xC4
#define MAP_TLV_METRIC_COLLECTION_INTERVAL             0xC5
#define MAP_TLV_RADIO_METRICS                          0xC6
#define MAP_TLV_AP_EXTENDED_METRICS                    0xC7
#define MAP_TLV_ASSOCIATED_STA_EXT_LINK_METRICS        0xC8
#define MAP_TLV_STATUS_CODE                            0xC9
#define MAP_TLV_REASON_CODE                            0xCA
#define MAP_TLV_BACKHAUL_STA_RADIO_CAPABILITY          0xCB
#define MAP_TLV_BACKHAUL_BSS_CONFIG                    0xD0


#define MULTIAP_PROFILE_1                              0x01
#define MULTIAP_PROFILE_2                              0x02

#define SUPPORTED_SERVICE_MULTIAP_CONTROLLER           0x00
#define SUPPORTED_SERVICE_MULTIAP_AGENT                0x01

#define SEARCHED_SERVICE_MULTIAP_CONTROLLER            0x00



#define IEEE1905_OBJECT_MULTIAP			"ieee1905.map"



typedef uint8_t macaddr_t[6];


/** TLV: AP Radio Identifier */
struct tlv_ap_radio_identifier {
	uint8_t radio[6];
} __attribute__((packed));


/** TLV: AP Operational BSS */
struct tlv_ap_oper_bss {
	uint8_t num_radio;
	struct ap_oper_bss_radiolist {
		uint8_t radio[6];
		uint8_t num_bss;
		struct ap_oper_bss_bss {
			uint8_t bssid[6];
			uint8_t ssidlen;
			char ssid[];
		} bss[];
	} radiolist[];
} __attribute__((packed));


/** TLV: Associated Clients */
struct tlv_assoc_client {
	uint8_t num_bss;
	struct assoc_client_bss {
		uint8_t bssid[6];
		uint16_t num_client;
		struct assoc_client_sta {
			uint8_t macaddr[6];
			uint16_t conntime;
		} sta[];
	} bss[];
} __attribute__((packed));


/** TLV: AP Capability */
struct tlv_ap_cap {
#define UNASSOC_STA_REPORTING_ONCHAN	0x80
#define UNASSOC_STA_REPORTING_OFFCHAN	0x40
#define AGENT_SUPPORTS_RCPI_STEER       0x20
	uint8_t cap;
} __attribute__((packed));

/** TLV: AP Radio Basic Capabilities */
struct tlv_ap_radio_basic_cap {
	uint8_t radio[6];
	uint8_t max_bssnum;
	uint8_t num_opclass;
	struct ap_radio_basic_cap_opclass {
		uint8_t classid;
		uint8_t max_txpower;
		uint8_t num_nonop_channel;
		uint8_t nonop_channel[];
	} __attribute__((packed)) opclass[];
} __attribute__((packed));

/** TLV: AP HT Capabilities */
struct tlv_ap_ht_cap {
	uint8_t radio[6];
	uint8_t cap;
#define HT_MAX_TX_STREAM_MASK	0xc0
#define HT_MAX_RX_STREAM_MASK	0x30
#define HT_SGI20_MASK		0x08
#define HT_SGI40_MASK		0x04
#define HT_HT40_MASK		0x02
} __attribute__((packed));

/** TLV: AP VHT Capabilities */
struct tlv_ap_vht_cap {
	uint8_t radio[6];
	uint16_t tx_mcs_supported;
	uint16_t rx_mcs_supported;
	uint8_t cap[2];
	/* cap[0] bitflags */
#define VHT_MAX_TX_STREAM_MASK	0xe0
#define VHT_MAX_RX_STREAM_MASK	0x1c
#define VHT_SGI80_MASK		0x02
#define VHT_SGI160_8080_MASK	0x01
	/* cap[1] bitflags */
#define VHT_8080_MASK		0x80
#define VHT_160_MASK		0x40
#define VHT_SU_BFR		0x20
#define VHT_MU_BFR		0x10
};

/** TLV: AP HE Capabilities */
struct ap_he_cap_mcs {
	uint8_t len;
	uint8_t mcs[];
} __attribute__((packed));

struct tlv_ap_he_cap {
	uint8_t radio[6];
	struct ap_he_cap_mcs hemcs;
	uint8_t cap[2];
	/* cap[0] bitflags */
#define HE_MAX_TX_STREAM_MASK	0xe0
#define HE_MAX_RX_STREAM_MASK	0x1c
#define HE_8080_MASK		0x02
#define HE_160_MASK		0x01
	/* cap[1] bitflags */
#define HE_SU_BFR		0x80
#define HE_MU_BFR		0x40
#define HE_UL_MUMIMO		0x20
#define HE_UL_MUMIMO_OFDMA	0x10
#define HE_DL_MUMIMO_OFDMA	0x08
#define HE_UL_OFDMA		0x04
#define HE_DL_OFDMA		0x02
} __attribute__((packed));


/** TLV: Steering Policy */
struct sta_macaddr {
	uint8_t num;
	uint8_t macaddr[][6];
} __attribute__((packed));

struct tlv_steering_policy {
	struct sta_macaddr nosteer;
	struct sta_macaddr nobtmsteer;
	uint8_t num_radio;
	struct {
		uint8_t radio[6];
		uint8_t steer_policy;
#define STEER_DISALLOW	0x00
#define STEER_MANDATE	0x01
#define STEER_ALLOW	0x02
		uint8_t util_threshold;
		uint8_t rcpi_threshold;
	} policy[];
} __attribute__((packed));


/** TLV: Metric Reporting Policy */
struct tlv_metric_report_policy {
	uint8_t interval;
	uint8_t num_radio;
	struct {
		uint8_t radio[6];
		uint8_t rcpi_threshold;
		uint8_t rcpi_hysteresis;
		uint8_t util_threshold;
#define INCLUDE_STA_STATS		0x80
#define INCLUDE_STA_LINK_METRICS	0x40
		uint8_t include;
	} policy[];
} __attribute__((packed));


/** TLV: Channel Preference */
struct channel_preflist {
	uint8_t num_channel;
	uint8_t channel[];
} __attribute__((packed));

struct tlv_channel_pref {
	uint8_t radio[6];
	uint8_t num_opclass;
	struct channel_pref_opclass {
		uint8_t classid;
		struct channel_preflist chs;
#define CHANNEL_PREF_MASK	0xf0
#define CHANNEL_PREF_REASON	0x0f
#define CHANNEL_PREF_REASON_UNSPEC			0x0
#define CHANNEL_PREF_REASON_NON11_INTERFERENCE		0x1
#define CHANNEL_PREF_REASON_INT_OBSS_INTERFERENCE	0x2
#define CHANNEL_PREF_REASON_EXT_OBSS_INTERFERENCE	0x3
#define CHANNEL_PREF_REASON_REDUCED_COVERAGE		0x5
#define CHANNEL_PREF_REASON_REDUCED_THROUGHPUT		0x6
#define CHANNEL_PREF_REASON_IN_DEVICE_INTERFERENCE	0x7
#define CHANNEL_PREF_REASON_DFS_NOP			0x8
#define CHANNEL_PREF_REASON_SHARED_BHAUL_PREVENT	0x9
#define CHANNEL_PREF_REASON_DFS_USABLE			0xa
#define CHANNEL_PREF_REASON_DFS_AVAILABLE		0xb
#define CHANNEL_PREF_REASON_REG_DISALLOWED		0xc
		uint8_t preference;
	} opclass[];
} __attribute__((packed));


/** TLV: Radio Operation Restriction */
struct channel_restrictlist {
	uint8_t num_channel;
	struct {
		uint8_t channel;
		uint8_t min_freq_sep;
	} channels[];
} __attribute__((packed));

struct tlv_radio_oper_restrict {
	uint8_t radio[6];
	uint8_t num_opclass;
	struct {
		uint8_t classid;
		struct channel_restrictlist chs;
	} opclass[];
} __attribute__((packed));


/** TLV: Transmit Power Limit */
struct tlv_txpower_limit {
	uint8_t radio[6];
	uint8_t limit;
} __attribute__((packed));

/** TLV: Channel Selection Response */
struct tlv_channel_selection_resp {
	uint8_t radio[6];
	uint8_t response;
} __attribute__((packed));


/** TLV: Operating Channel Report */
struct oper_channel_report {
	uint8_t num_opclass;
	struct {
		uint8_t classid;
		uint8_t channel;
	} opclass[];
} __attribute__((packed));

struct tlv_oper_channel_report {
	uint8_t radio[6];
	struct oper_channel_report report;
	uint8_t curr_txpower;
} __attribute__((packed));

/** TLV: Client Info */
struct tlv_client_info {
	uint8_t bssid[6];
	uint8_t macaddr[6];
} __attribute__((packed));

/** TLV: Client Capability Report */
struct tlv_client_cap_report {
	uint8_t result;
	uint8_t frame[];
} __attribute__((packed));


/** TLV: Client Association Event */
struct tlv_client_assoc_event {
	uint8_t macaddr[6];
	uint8_t bssid[6];
#define CLIENT_EVENT_MASK	0x80
	uint8_t event;
} __attribute__((packed));


/** TLV: AP Metric Query */
struct tlv_ap_metric_query {
	uint8_t num_bss;
	struct {
		uint8_t bssid[6];
	} bss[];
} __attribute__((packed));


/** TLV: AP Metrics TLV */
struct tlv_ap_metrics {
	uint8_t bssid[6];
	uint8_t channel_utilization;
	uint16_t num_station;
#define ESP_AC_BE	0x80
#define ESP_AC_BK	0x40
#define ESP_AC_VO	0x20
#define ESP_AC_VI	0x10
	uint8_t esp_ac;
	uint8_t esp_be[3];
	uint8_t esp[];
} __attribute__((packed));


/** TLV: STA MAC Address Type */
struct tlv_sta_mac {
	uint8_t macaddr[6];
} __attribute__((packed));


/** TLV: Associated STA Link Metrics */
struct assoc_sta_link_metrics_bss {
	uint8_t bssid[6];
	uint32_t time_delta;
	uint32_t dl_thput;
	uint32_t ul_thput;
	uint8_t ul_rcpi;
} __attribute__((packed));

struct tlv_assoc_sta_link_metrics {
	uint8_t macaddr[6];
	uint8_t num_bss;
	struct assoc_sta_link_metrics_bss bss[];
} __attribute__((packed));

/** TLV: Unssociated STA Link Metrics Query */
struct tlv_unassoc_sta_link_metrics_query {
	uint8_t opclass;
	uint8_t num_channel;
	struct {
		uint8_t channel;
		uint8_t num_sta;
		struct {
			uint8_t macaddr[6];
		} sta[];
	} ch[];
} __attribute__((packed));

/** TLV: Unssociated STA Link Metrics Response */
struct unassoc_sta_link_metrics_sta {
	uint8_t macaddr[6];
	uint8_t channel;
	uint32_t time_delta;
	uint8_t ul_rcpi;
} __attribute__((packed));

struct tlv_unassoc_sta_link_metrics_resp {
	uint8_t opclass;
	uint8_t num_sta;
	struct unassoc_sta_link_metrics_sta sta[];
} __attribute__((packed));


/** TLV: Beacon Metrics Query */
struct ssid_query {
	uint8_t ssidlen;
	char ssid[];
} __attribute__((packed));

struct ap_channel_report {
	uint8_t len;
	uint8_t opclass;
	uint8_t channel[];
} __attribute__((packed));

struct tlv_beacon_metrics_query {
	uint8_t sta_macaddr[6];
	uint8_t opclass;
	uint8_t channel;
	uint8_t bssid[6];
	uint8_t reporting_detail;
	struct ssid_query ssid;
	uint8_t num_report;
	struct ap_channel_report report;
	uint8_t num_element;
	uint8_t element[];
} __attribute__((packed));


/** TLV: Beacon Metrics Response */
struct tlv_beacon_metrics_resp {
	uint8_t sta_macaddr[6];
	uint8_t reserved;
	uint8_t num_element;
	uint8_t element[];
} __attribute__((packed));

/** TLV: Steering Request */
struct steer_target_bss {
	uint8_t num;
	struct {
		uint8_t bssid[6];
		uint8_t opclass;
		uint8_t channel;
	} bss[];
} __attribute__((packed));

struct steer_sta_request {
	uint8_t num;
	struct {
		uint8_t macaddr[6];
	} sta[];
} __attribute__((packed));

struct tlv_steer_request {
	uint8_t bssid[6];
	uint8_t mode;
#define STEER_REQUEST_MODE		0x80
#define STEER_REQUEST_BTM_DISASSOC_IMM	0x40
#define STEER_REQUEST_BTM_ABRIDGED	0x20
	uint16_t op_window;
	uint16_t btm_disassoc_timer;
	struct steer_sta_request sta;
	struct steer_target_bss target;
} __attribute__((packed));


/** TLV: Steering BTM Report */
struct tlv_steer_btm_report {
	uint8_t bssid[6];
	uint8_t sta_macaddr[6];
	uint8_t status;
	macaddr_t target_bssid[];
} __attribute__((packed));


/** TLV: Client Association Control Request */
struct tlv_client_assoc_ctrl_request {
	uint8_t bssid[6];
	uint8_t control;
#define ASSOC_CTRL_BLOCK	0x00
#define ASSOC_CTRL_UNBLOCK	0x01
	uint16_t validity_period;
	uint8_t num_sta;
	struct {
		uint8_t macaddr[6];
	} sta[];
} __attribute__((packed));


/** TLV: Backhaul Steering Request */
struct tlv_backhaul_steer_request {
	uint8_t macaddr[6];
	uint8_t target_bssid[6];
	uint8_t target_opclass;
	uint8_t target_channel;
} __attribute__((packed));

/** TLV: Backhaul Steering Response */
struct tlv_backhaul_steer_resp {
	uint8_t macaddr[6];
	uint8_t target_bssid[6];
	uint8_t result;
} __attribute__((packed));


/** TLV: Higher Layer Data */
struct tlv_higher_layer_data {
	uint8_t protocol;
	uint8_t payload[];
} __attribute__((packed));


/** TLV: Associated STA Traffic Stats */
struct tlv_assoc_sta_traffic_stats {
	uint8_t macaddr[6];
	uint32_t tx_bytes;
	uint32_t rx_bytes;
	uint32_t tx_packets;
	uint32_t rx_packets;
	uint32_t tx_err_packets;
	uint32_t rx_err_packets;
	uint32_t rtx_packets;
} __attribute__((packed));


/** TLV: Error code */
struct tlv_error_code {
	uint8_t reason;
	uint8_t macaddr[6];
} __attribute__((packed));


/** TLV: Channel scan reporting policy */
struct tlv_channel_scan_report_policy {
	uint8_t report;
#define REPORT_CHANNEL_SCANS	0x80
} __attribute__((packed));


/** TLV: Channel scan capabilities */
struct channel_scan_capability_opclass {
	uint8_t classid;
	uint8_t num_channel;
	uint8_t channel[];
} __attribute__((packed));

struct channel_scan_capability_radio {
	uint8_t radio[6];
	uint8_t cap;
#define SCAN_CAP_ON_BOOT_ONLY	0x80
#define SCAN_CAP_IMPACT		0x60
	uint32_t min_scan_interval;
	uint8_t num_opclass;
	struct channel_scan_capability_opclass opclass[];
} __attribute__((packed));

struct tlv_channel_scan_capability {
	uint8_t num_radio;
	struct channel_scan_capability_radio radio[];
} __attribute__((packed));


/** TLV: Channel scan request */
struct tlv_channel_scan_request {
	uint8_t mode;
#define SCAN_REQUEST_FRESH_SCAN	0x80
	uint8_t num_radio;
	struct channel_scan_request_radio {
		uint8_t radio[6];
		uint8_t num_opclass;
		struct channel_scan_request_opclass {
			uint8_t classid;
			uint8_t num_channel;
			uint8_t channel[];
		} opclass[];
	} radio[];
} __attribute__((packed));


/** TLV: Channel scan result */
struct scan_result_timestamp {
	uint8_t len;
	char timestamp[];
} __attribute__((packed));

struct scan_result_ssid {
	uint8_t len;
	uint8_t ssid[];
} __attribute__((packed));

struct scan_result_bandwidth {
	uint8_t len;
	char bwstr[];		/* weird definition! */
} __attribute__((packed));

struct scan_result_bssload_data {
	uint8_t ch_util;
	uint16_t sta_count;
} __attribute__((packed));

struct scan_result_bssload {
	uint8_t info;
#define CH_SCAN_RESULT_BSSLOAD_PRESENT		0x80
	struct scan_result_bssload_data data[];
} __attribute__((packed));

struct scan_result_neighbor {
	uint16_t num_neighbor;
	struct {
		uint8_t bssid[6];
		struct scan_result_ssid ssid;
		uint8_t rcpi;
		struct scan_result_bandwidth bw;
		struct scan_result_bssload bssload;
	} neighbor[];
} __attribute__((packed));

struct scan_result_detail {
	struct scan_result_timestamp tsp;
	uint8_t utilization;
	uint8_t noise;
	struct scan_result_neighbor nbr;
	uint32_t total_scan_duration;
#define SCAN_RESULT_SCAN_TYPE	0x80
	uint8_t type;
} __attribute__((packed));

struct tlv_channel_scan_result {
	uint8_t radio[6];
	uint8_t opclass;
	uint8_t channel;
#define CH_SCAN_STATUS_SUCCESS                  0x00
#define CH_SCAN_STATUS_SCAN_NOT_SUPPORTED       0x01
#define CH_SCAN_STATUS_TOO_SOON                 0x02
#define CH_SCAN_STATUS_TOO_BUSY                 0x03
#define CH_SCAN_STATUS_SCAN_NOT_COMPLETED       0x04
#define CH_SCAN_STATUS_SCAN_ABORTED             0x05
#define CH_SCAN_STATUS_BOOT_SCAN_ONLY           0x06
	uint8_t status;
	struct scan_result_detail detail[];	/* present when satus = 0 */
} __attribute__((packed));


/** TLV: Timestamp */
struct tlv_timestamp {
	uint8_t len;
	uint8_t timestamp[];
} __attribute__((packed));


/** TLV: CAC request */
struct tlv_cac_request {
	uint8_t num_radio;
	struct {
		uint8_t radio[6];
		uint8_t opclass;
		uint8_t channel;
#define CAC_REQUEST_METHOD		0xe0
#define CAC_REQUEST_COMPLETE_ACTION	0x18
		uint8_t mode;
	} radio[];
} __attribute__((packed));


/** TLV: CAC termination */
struct tlv_cac_termination {
	uint8_t num_radio;
	struct {
		uint8_t radio[6];
		uint8_t opclass;
		uint8_t channel;
	} radio[];
} __attribute__((packed));


/** TLV: CAC completion report */
struct tlv_cac_complete_report {
	uint8_t num_radio;
	struct {
		uint8_t radio[6];
		uint8_t opclass;
		uint8_t channel;
#define CAC_COMP_REPORT_STATUS_SUCCESSFUL                       0x00
#define CAC_COMP_REPORT_STATUS_RADAR_DETECTED                   0x01
#define CAC_COMP_REPORT_STATUS_CAC_NOT_SUPPORTED                0x02
#define CAC_COMP_REPORT_STATUS_TOO_BUSY                         0x03
#define CAC_COMP_REPORT_STATUS_NON_CONFORMANT                   0x04
#define CAC_COMP_REPORT_STATUS_OTHER                            0x05
		uint8_t status;
		uint8_t num_pairs;	/* radars detected in pairs */
		struct {
			uint8_t opclass;
			uint8_t channel;
		} pair[];
	} radio[];
} __attribute__((packed));


/** TLV: CAC status report */
struct cac_status_available {
	uint8_t num_channel;
	struct {
		uint8_t opclass;
		uint8_t channel;
		uint16_t since;		/* in minutes */
	} ch[];
} __attribute__((packed));

struct cac_status_noop {
	uint8_t num_channel;
	struct {
		uint8_t opclass;
		uint8_t channel;
		uint16_t remain;	/* in seconds */
	} ch[];
} __attribute__((packed));

struct cac_status_cac {
	uint8_t num_channel;
	struct {
		uint8_t opclass;
		uint8_t channel;
		uint16_t remain;	/* in seconds */
	} ch[];
} __attribute__((packed));

struct tlv_cac_status_report {
	struct cac_status_available available;
	struct cac_status_noop noop;
	struct cac_status_cac cac;
} __attribute__((packed));


/** TLV: CAC capability */
struct tlv_cac_cap {
	uint8_t country[2];
	uint8_t num_radio;
	struct cac_cap_radio {
		uint8_t radio[6];
		uint8_t num_cac;
		struct cac_cap_cac {
#define CAC_METHOD_CONTINUOUS_CAC	0x00
#define CAC_METHOD_DEDICATED_RADIO	0x01
#define CAC_METHOD_MIMO_DIM_REDUCED     0x02
#define CAC_METHOD_TIME_SLICED		0x03
			uint8_t supp_method;
			uint8_t duration[3];
			uint8_t num_opclass;
			struct cac_cap_opclass {
				uint8_t classid;
				uint8_t num_channel;
				uint8_t channel[];
			} __attribute__((packed)) opclass[];
		} __attribute__((packed)) cac[];
	} __attribute__((packed)) radio[];
} __attribute__((packed));


/** TLV: Multi-AP profile */
struct tlv_map_profile {
	uint8_t profile;
} __attribute__((packed));


/** TLV: Profile-2 AP capability */
struct tlv_profile2_ap_cap {
	uint16_t reserved;
#define STATS_UNIT_MASK		0x60
#define STATS_UNIT_BYTE		0x00
#define STATS_UNIT_KB		0x40
#define STATS_UNIT_MB		0x80
	uint8_t unit;
	uint8_t max_vids;
} __attribute__((packed));


/** TLV: Default 802.1Q settings */
struct tlv_default_8021q_settings {
	uint16_t pvid;
	uint8_t pcp;
#define PCP_MASK  0xe0
} __attribute__((packed));

/** TLV: Traffic separation policy */
struct ssid_info {
	uint8_t len;
	uint8_t ssid[];
} __attribute__((packed));

struct tlv_traffic_sep_policy {
	uint8_t num_ssid;
	struct {
		struct ssid_info info;
		uint16_t vid;
	} __attribute__((packed)) ssid[];
} __attribute__((packed));


/** TLV: Profile-2 error code */
struct tlv_profile2_error_code {
	uint8_t reason;
	macaddr_t bssid[];
} __attribute__((packed));


/** TLV: AP radio advanced capability */
struct tlv_ap_radio_adv_cap {
	uint8_t radio[6];
	uint8_t cap;
#define RADIO_CAP_ADVANCED_MASK		0x60
#define RADIO_CAP_COMBINED_FHBK		0x80
#define RADIO_CAP_COMBINED_P1P2		0x40
} __attribute__((packed));


/** TLV: Association status notification */
struct tlv_assoc_status_notif {
	uint8_t num_bss;
	struct {
		uint8_t bssid[6];
		uint8_t allowance;
#define STA_ASSOC_NOT_ALLOWED	0x00
#define STA_ASSOC_ALLOWED	0x01
	} bss[];
} __attribute__((packed));


/** TLV: Source info */
struct tlv_source_info {
	uint8_t macaddr[6];
} __attribute__((packed));

/** TLV: Tunneled message type */
struct tlv_tunnel_msg_type {
#define TUN_MSG_ASSOC_REQ	0x00
#define TUN_MSG_REASSOC_REQ	0x01
#define TUN_MSG_BTM_QUERY	0x02
#define TUN_MSG_WNM_REQ		0x03
#define TUN_MSG_ANQP_NBR_RPT	0x04
	uint8_t type;
} __attribute__((packed));


/** TLV: Tunneled message */
struct tlv_tunneled {
	uint8_t frame[0];
} __attribute__((packed));


/** TLV: Profile-2 steering request */
struct profile2_sta_steer {
	uint8_t num_sta;
	struct {
		uint8_t macaddr[6];
	} sta[];
} __attribute__((packed));

struct tlv_profile2_steer_request {
	uint8_t bssid[6];
	uint8_t mode;
#define STEER_MODE		0x80
#define STEER_BTM_DISASSOC_IMM	0x40
#define STEER_BTM_ABRIDGED	0x20
	uint16_t op_window;
	uint16_t btm_disassoc_timer;
	struct profile2_sta_steer steer;
	uint8_t num_target;
	struct {
		uint8_t bssid[6];
		uint8_t opclass;
		uint8_t channel;
		uint8_t reason;
	} target[];
} __attribute__((packed));

/** TLV: Unsuccessful association policy */
struct tlv_unsuccess_assoc_policy {
	uint8_t report;
#define UNSUCCESSFUL_ASSOC_REPORT  0x80
	uint32_t max_report_rate;
} __attribute__((packed));

/** TLV: Metric collection interval */
struct tlv_metric_collection_int {
	uint32_t interval;
} __attribute__((packed));


/** TLV: Radio metrics */
struct tlv_radio_metrics {
	uint8_t radio[6];
	uint8_t noise;
	uint8_t transmit;
	uint8_t receive_self;
	uint8_t receive_other;
} __attribute__((packed));


/** TLV: AP extended metrics */
struct tlv_ap_ext_metrics {
	uint8_t bssid[6];
	uint32_t tx_bytes_ucast;
	uint32_t rx_bytes_ucast;
	uint32_t tx_bytes_mcast;
	uint32_t rx_bytes_mcast;
	uint32_t tx_bytes_bcast;
	uint32_t rx_bytes_bcast;
} __attribute__((packed));


/** TLV: Associated STA extended link metrics */
struct sta_ext_link_metric_bss {
	uint8_t bssid[6];
	uint32_t dl_rate;
	uint32_t ul_rate;
	uint32_t rx_util;
	uint32_t tx_util;
} __attribute__((packed));

struct tlv_sta_ext_link_metric {
	uint8_t macaddr[6];
	uint8_t num_bss;
	struct sta_ext_link_metric_bss bss[];
} __attribute__((packed));

/** TLV: Status code */
struct tlv_status_code {
	uint16_t code;
} __attribute__((packed));

/** TLV: Reason code */
struct tlv_reason_code {
	uint16_t code;
} __attribute__((packed));


/** TLV: Backhaul STA radio capabilities */
struct tlv_bsta_radio_cap {
	uint8_t radio[6];
	uint8_t macaddr_included;
#define BSTA_MACADDRESS_INCLUDED	0x80
	macaddr_t macaddr[];
} __attribute__((packed));


/** TLV: Backhaul BSS config */
struct tlv_bbss_config {
	uint8_t bssid[6];
	uint8_t config;
#define BBSS_CONFIG_P1_BSTA_DISALLOWED	0x80
#define BBSS_CONFIG_P2_BSTA_DISALLOWED	0x40
} __attribute__((packed));


#endif /* MAP2_H */
